using System;
using System.Net;
using System.Threading;
using System.Net.Sockets;

using NetSystem.Server.Collections;

namespace NetSystem.Server
{
	internal delegate void ReceivedDataFromClient(NetPacket thePacket, NetSystem.Server.Client theClient);

	public delegate void ReceivedData(int theClientID, NetPacket thePacket);
	public delegate void NewConnection(int theClientID, IPAddress theClientIP);
	public delegate void ClientDisconnected(int theClientID, string theReason);

	public class NetServer
	{
		private Socket m_ServerSocket;
		private ClientCollection m_ClientPool;
		private IPAddressCollection m_IPBanList;

		public event ReceivedData EventReceivedData;
		public event NewConnection EventNewConnection;
		public event ClientDisconnected EventClientDisconnected;

		public NetServer(int thePort)
		{
			m_ServerSocket	= new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
			m_ClientPool	= new ClientCollection();
			m_IPBanList		= new IPAddressCollection();

			m_ServerSocket.Bind(new IPEndPoint(IPAddress.Any, thePort));
		}

		#region Public Methods
		public void StartListen()
		{
			try
			{
				m_ServerSocket.Listen(10);
				m_ServerSocket.BeginAccept(new AsyncCallback(AcceptConnection), this);
			}
			catch(SocketException ex)
			{
				throw ex;
			}
		}

		public void StopListen()
		{
			m_ServerSocket.Shutdown(SocketShutdown.Both);
		}

		public void SendData(int theClientID, NetPacket thePacket)
		{
			Client client = m_ClientPool.GetClientByID(theClientID);
			client.SendData(thePacket);
		}

		public void ClientDisconnect(int theClientID, string theReason)
		{
			Client c = m_ClientPool.GetClientByID(theClientID);
			c.SendData(new NetDisconnect(theReason));

			m_ClientPool.Remove(c);
			c.Close();
		}

		public void Close()
		{
			m_ServerSocket.Shutdown(SocketShutdown.Both);
			m_ServerSocket.Close();
		}
		#endregion

		#region Private Methods
		private void AcceptConnection(IAsyncResult theAR)
		{
			NetServer ns = (NetServer)theAR.AsyncState;
			newConnection(ns.m_ServerSocket.EndAccept(theAR));
			ns.m_ServerSocket.BeginAccept(new AsyncCallback(AcceptConnection), ns);
		}

		private void newConnection(Socket theSocket)
		{
			int cId = m_ClientPool.GetFreeClientId();
			Client c = new Client(theSocket, cId);
			c.EventDisconnect += new NetSystem.Server.Client.ClientDisconnect(client_EventDisconnect);
			c.EventReceivedData += new ReceivedDataFromClient(client_EventReceivedData);

			if(m_IPBanList.IndexOf(c.ClientIP) > -1)
			{
				c.SendData(new NetDisconnect("You are banned!"));
			}
			else
			{
				Monitor.Enter(m_ClientPool);
				m_ClientPool.Add(c);
				Monitor.Exit(m_ClientPool);
				c.StartReading();

				EventNewConnection(c.ClientID, c.ClientIP);
			}
		}
		#endregion

		#region Event Handler
		private void client_EventReceivedData(NetPacket thePacket, Client theClient)
		{
			EventReceivedData(theClient.ClientID, thePacket);
		}

		private void client_EventDisconnect(Client theClient, string theReason)
		{
			m_ClientPool.Remove(theClient);
			EventClientDisconnected(theClient.ClientID, theReason);
			theClient.Close();			
		}
		#endregion

		public IPAddressCollection IPBanList
		{
			get { return m_IPBanList; }
		}
	}
}